/* $Header: /src/carson/nistnet/RCS/knistnet.c,v 1.7 1998/06/04 18:29:22 carson Exp $ */

/* knistnet.c - Linux implementation of "hitbox"-like functionality.
 * This code exists as a loadable kernel module.  It gains access to the
 * entry points it needs through some patches to the existing Linux kernel.
 * (Unfortunately, there seemed to be no practical alternative to patching
 * to gain access to the basic packet handling routine entry points.)
 * The user-level interface provided is a device driver one, modeled on the
 * original SunOS-based hitbox.
 */

#include "kincludes.h"
#include <linux/version.h>
#include <linux/module.h>
#include <net/checksum.h>

#include "tabledist.h"

int domunge;
int hit_debug;

/* Breakpoints are only helpful when we're compiled into the kernel */
#ifndef MODULE
#ifdef DEBUG
#define BREAKPOINT() asm("   int $3")
#else
#define BREAKPOINT()
#endif
#else
#define BREAKPOINT()
#endif

#include <asm/bitops.h>
#define LOCK_BIT 1
int LinLockVar;

#define LinLock() \
		{\
		save_flags(pre_flags);\
		cli();\
		if (set_bit(LOCK_BIT, &LinLockVar)) {\
			BREAKPOINT();\
		}\
		}
#define LinUnlock() \
		{\
		if (!clear_bit(LOCK_BIT, &LinLockVar)) {\
			BREAKPOINT();\
		}\
		restore_flags(pre_flags);\
		}


#define HASHSIZE 256
static struct lin_hitbox *hittable[HASHSIZE];

#ifdef DEBUG
static int hittable_count;
#endif

struct lin_globalstats ourstats;
#define	STATS_START	0
#define	STATS_PROCESS	1
#define STATS_UNPROCESS	2

void fixed_gettimeofday(struct timeval *tv);
void lin_hash_stats(int number);

void fast_fill(void);
void fast_empty(void);
struct fast_timer_list * fast_alloc(int how);
void fast_free(struct fast_timer_list *done);

void
freehittable(void)
{
	struct lin_hitbox *freeme, *frome;
	int i;
#ifdef DEBUG
	int count=0;
#endif

	for (i=0; i < HASHSIZE; ++i) {
		freeme = hittable[i];
		while (freeme) {
			frome = freeme->next;
			kfree_s(freeme, sizeof(struct lin_hitbox));
			freeme = frome;
#ifdef DEBUG
			++count;
#endif
		}
	}
#ifdef DEBUG
	if (count != hittable_count) {
		printk("nistnet: expected %d in hittable, found %d\n",
			hittable_count, count);
	}
#endif
}


/* Hash an entry by src/dest address */
int
hithash(u_int32_t src, u_int32_t dest)
{
	unsigned int value, answer=0;

	/* this is not intended to be an extraordinarily good hash! */
	value = (unsigned int)src*69069 + (unsigned int)dest;
	while (value) {
		answer |= value;
		value >>= 8;
	}
	return (int) (answer&(HASHSIZE-1));
}

struct lin_hitbox *
findhit(u_int32_t src, u_int32_t dest)
{
	register struct lin_hitbox *answer;
	int hash_tries = 0;
	
	answer = hittable[hithash(src, dest)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == src
			&& answer->hit_entry.dest == dest) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}

	++hash_tries;
	answer = hittable[hithash(src, INADDR_ANY)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == src
			&& answer->hit_entry.dest == INADDR_ANY) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}

	++hash_tries;
	/* Try again with generic addresses */
	answer = hittable[hithash(INADDR_ANY, dest)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == INADDR_ANY
			&& answer->hit_entry.dest == dest) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}
if (hit_debug) {
	unsigned char *srcp, *destp;

	srcp = (unsigned char *)&src;
	destp = (unsigned char *)&dest;
	printk("nistnet: missed on %d.%d.%d.%d->%d.%d.%d.%d\n",
			srcp[0], srcp[1], srcp[2], srcp[3],
			destp[0], destp[1], destp[2], destp[3]);
}
	++hash_tries;
	answer = hittable[hithash(INADDR_ANY, INADDR_ANY)];
	while (answer) {
		++hash_tries;
		if (answer->hit_entry.src == INADDR_ANY
			&& answer->hit_entry.dest == INADDR_ANY) {
			lin_hash_stats(hash_tries);
			return answer;
		}
		answer = answer->next;
	}
	++hash_tries;
	lin_hash_stats(hash_tries);
	return NULL;
}

int
rmhitreq(struct lin_hitreq *hitreq)
{
	struct lin_hitbox *victim, *parent;
	
	victim = hittable[hithash(hitreq->src, hitreq->dest)];
	if (!victim) return -ESRCH;
	if (victim->hit_entry.src == hitreq->src &&
		victim->hit_entry.dest == hitreq->dest) {
			hittable[hithash(hitreq->src, hitreq->dest)]
						= victim->next;
			kfree_s(victim, sizeof(struct lin_hitbox));
#ifdef DEBUG
			--hittable_count;
#endif
	} else for (parent=victim, victim=victim->next; victim;
			parent=victim, victim=victim->next) {
		if (victim->hit_entry.src == hitreq->src &&
			victim->hit_entry.dest == hitreq->dest) {
			parent->next = victim->next;
			kfree_s(victim, sizeof(struct lin_hitbox));
#ifdef DEBUG
			--hittable_count;
#endif
			break;
		}
	}
	return 0;
}


int
addhitreq(struct lin_hitreq *hitreq)
{
	struct lin_hitbox *boxer;

	/* Check entry for sanity */
	if (hitreq->drd_min > hitreq->drd_max)
		return -EINVAL;
#ifdef CONFIG_ECN
	if (hitreq->drd_congestion) {
		if (hitreq->drd_min > hitreq->drd_congestion ||
			hitreq->drd_max < hitreq->drd_congestion)
				return -EINVAL;
	}
#else
	hitreq->drd_congestion = 0;
#endif
	/* Change delays from usec to minijiffies */
	hitreq->delay = usec_to_minijiffy(hitreq->delay);
	hitreq->delsigma = usec_to_minijiffy(hitreq->delsigma);


	/* First see if we're already hitting on this one */
	boxer = findhit(hitreq->src, hitreq->dest);
	if (boxer && boxer->hit_entry.src == hitreq->src &&
		boxer->hit_entry.dest == hitreq->dest) {
		/* replace entry */
		boxer->hit_entry = *hitreq;
	} else {
		/* need a new one */
		boxer = kmalloc(sizeof(struct lin_hitbox), GFP_KERNEL);
		if (!boxer)
			return -ENOMEM;
		memset(boxer, 0, sizeof(struct lin_hitbox));
		boxer->hit_entry = *hitreq;
		/* Initialize packet timer */
		fixed_gettimeofday(&boxer->stats.last_packet);
		boxer->stats.next_packet = boxer->stats.last_packet;
		/* Insert in table */
		boxer->next = hittable[hithash(hitreq->src, hitreq->dest)];
		hittable[hithash(hitreq->src, hitreq->dest)] = boxer;
#ifdef DEBUG
			++hittable_count;
#endif
	}
	return 0;
}


/* nice and high, but it is tunable: insmod hitmod.o major=your_selection */
static int major = HITMAJOR;

/*
 * The driver.
 */

/* read -- get what's in the table */
static int hw_read(struct inode * node,struct file * file,char * buf,int count)
{
	int i, origcount = count;
	struct lin_hitbox *readme;
	struct addrpair out;

	for (i=0; i < HASHSIZE && count >= sizeof(struct addrpair); ++i) {
		readme = hittable[i];
		while (readme && count >= sizeof(struct addrpair)) {

			out.src = readme->hit_entry.src;
			out.dest = readme->hit_entry.dest;
			memcpy_tofs(buf, (char *)&out,
				(unsigned long) sizeof(struct addrpair));
			readme = readme->next;
			buf += sizeof(struct addrpair);
			count -= sizeof(struct addrpair);
		}
	}
#ifdef DEBUG
	if (origcount-count > hittable_count) {
		printk("nistnet: %d new guys showed up in hittable somehow\n",
			(origcount-count)-hittable_count);
	}
#endif
	return origcount-count;
}

static int hw_write(struct inode * node,struct file * file,const char * buf,int count)
{
	if (count >= tabledistsize()) {
		tabledistfill(buf);
		return tabledistsize();
	} else
		return -E2BIG;	/* ha, ha */
}

/*
 * mucky muck ioctl interface
 */
static int hw_ioctl(struct inode * inode, struct file * file, unsigned int type, unsigned long arg)
{
	struct lin_hitreq hitreq;
	int rc = 0;

	switch (type) {
	case HITIOCTL_OFF:
		/* Shut down */
		domunge=0;
		break;
	case HITIOCTL_ON:
		/* Turn on if not already on */
		domunge=1;
		break;
	case HITIOCTL_ADD:
		/* Get what they want */
		memcpy_fromfs(&hitreq, (struct lin_hitreq *)arg,
			sizeof(struct lin_hitreq));
		/* Add it to the table */
		rc = addhitreq(&hitreq);
		break;
	case HITIOCTL_REMOVE:
		/* Get what they want */
		memcpy_fromfs(&hitreq, (struct lin_hitreq *)arg,
			sizeof(struct lin_hitreq));
		/* Remove it from the table */
		rc = rmhitreq(&hitreq);
		break;
	case HITIOCTL_STATS:
		{
		struct lin_hitstats hitstats;
		struct lin_hitbox *where;
		int i;

		/* Get what they want */
		memcpy_fromfs(&hitstats, (struct lin_hitstats *)arg,
			sizeof(struct lin_hitstats));
		where = findhit(hitstats.hitreq.src, hitstats.hitreq.dest);
		if (!where)
			return -ESRCH;
		hitstats = where->stats;
		/* Change delays from minijiffies to usec */
		hitstats.hitreq.delay = minijiffy_to_usec(hitstats.hitreq.delay);
		hitstats.hitreq.delsigma = minijiffy_to_usec(hitstats.hitreq.delsigma);
		/* Compute current_bandwidth */
		for (i=0; i < BAND_ARRAY; ++i)
			hitstats.current_bandwidth +=
				hitstats.bandwidth_array[i];
		if (hitstats.seats_used)
			hitstats.current_bandwidth /= hitstats.seats_used;
		/* Copy out individual stats */
		memcpy_tofs((struct lin_hitstats *)arg,
			&hitstats,
			sizeof(struct lin_hitstats));
		break;
		}
	case HITIOCTL_MODE:
		/* ?? */
		break;
	case HITIOCTL_TIMER:
		/* ?? */
		break;
	case HITIOCTL_MTU:
		/* ?? */
		break;
	case HITIOCTL_DEBUG:
		hit_debug = get_user((long *)arg);
		break;
	case HITIOCTL_GLOBALSTATS:
		memcpy_tofs((struct lin_globalstats *)arg,
			&ourstats,
			sizeof(struct lin_globalstats));
		break;
	}
	return rc;
}


/*
 * Our special open code.
 * MOD_INC_USE_COUNT make sure that the driver memory is not freed
 * while the device is in use.
 */
static int
hw_open( struct inode* ino, struct file* filep)
{
	MOD_INC_USE_COUNT;
	return 0;   
}

/*
 * Now decrement the use count.
 */
static void
hw_close( struct inode* ino, struct file* filep)
{
	MOD_DEC_USE_COUNT;
}

static struct file_operations hw_fops = {
	NULL,
	hw_read,
	hw_write,
	NULL,		/* hw_readdir */
	NULL,		/* hw_select */
	hw_ioctl,		/* hw_ioctl */
	NULL,
	hw_open,
	hw_close,
	NULL		/* fsync */
};


/* Time routines */


inline unsigned long inabs(long a)
{
	return a >= 0 ? a : -a;
}

/* return a-b difference in microseconds */
static long
timeval_diff(struct timeval *a, struct timeval *b)
{
	long int answer;

	if (a->tv_sec > b->tv_sec) {
		answer = (a->tv_sec - b->tv_sec)-1;
		answer *= MILLION;
		answer += a->tv_usec+MILLION - b->tv_usec;
		return answer;
	} else if (a->tv_sec == b->tv_sec) {
		return a->tv_usec - b->tv_usec;
	} else {
		answer = (b->tv_sec - a->tv_sec)-1;
		answer *= MILLION;
		answer += b->tv_usec+MILLION - a->tv_usec;
		return -answer;
	}
}

/* Return time difference a-b in minijiffies */
static int
timeval_minidiff(struct timeval *a, struct timeval *b)
{
	return usec_to_minijiffy(timeval_diff(a, b));
}

/* Add on some usec to time */
static void
timeval_add(struct timeval *a, unsigned long more_usec)
{
	a->tv_usec += more_usec;
	if (a->tv_usec > MILLION) {
		a->tv_usec -= MILLION;
		++a->tv_sec;
	}
}

/* This version has some elementary hacks to try to cope with minor
 * time warp problems.  With the "CMT" fixes to the clock, though,
 * it's highly unlikely any such problems will occur.
 */
void
fixed_gettimeofday(struct timeval *tv)
{
	static struct timeval last;

	do_gettimeofday(tv);
	/* If it looks like the time went backwards, try again... */
	if (tv->tv_sec < last.tv_sec ||
		(tv->tv_sec == last.tv_sec && tv->tv_usec < last.tv_usec)) {
		do_gettimeofday(tv);
	}
	/* If it's still backwards, then what?  The problem could be either
	 * with this value or the previous value...
	 */
	if (tv->tv_sec < last.tv_sec ||
	    (tv->tv_sec == last.tv_sec && tv->tv_usec < last.tv_usec)) {
#ifdef TIME_DEBUG
		static struct timeval previous;

		if (previous.tv_usec != last.tv_usec)
			printk("nistnet: clock underflow at %ld.%06ld from %ld.%06ld! Went back %ld usecs\n",
			tv->tv_sec, tv->tv_usec,
			last.tv_sec, last.tv_usec,
			timeval_diff(&last, tv));
		previous = last;
#endif
		*tv = last;
	}
	last = *tv;
}

/* Various statistics.  We record some running totals in circular arrays,
 * to give an idea of how things are going.
 */
#define band_seat(time) ((time.tv_sec)%BAND_ARRAY)


/* How well is the hash table doing? */
void
lin_hash_stats(int number)
{
	static int hash_slot;

	/* average 50/50 with the last value we got */
	ourstats.hash_tries[hash_slot] = (number + ourstats.hash_tries[hash_slot]) >> 1;
	hash_slot = (hash_slot+1)%BAND_ARRAY;
}

/* How long is our processing time? */
void
global_stats(int process)
{
	static struct timeval start;
	struct timeval end;
	static int process_slot, unprocess_slot;
	long int usec_time;

	switch(process) {
	case STATS_START:
		fixed_gettimeofday(&start);
		return;
	case STATS_PROCESS:
		fixed_gettimeofday(&end);
		usec_time = timeval_diff(&end, &start);
#ifdef TIME_DEBUG
		/* Check for bogus time values */
		if (usec_time < 0) {
			printk("nistnet: pretty fast processing, %ld usecs!\n", usec_time);
			return;
		}
#endif
		/* Average in with previous value 50/50 */
		ourstats.process_overhead[process_slot] =
			(usec_time+ourstats.process_overhead[process_slot])>> 1;
		process_slot = (process_slot+1)%BAND_ARRAY;
		break;
	case STATS_UNPROCESS:
		fixed_gettimeofday(&end);
		usec_time = timeval_diff(&end, &start);
#ifdef TIME_DEBUG
		/* Check for bogus time values */
		if (usec_time < 0) {
			printk("nistnet: pretty fast unprocessing, %ld usecs!\n", usec_time);
			return;
		}
#endif
		/* Average in with previous value 50/50 */
		ourstats.unprocess_overhead[unprocess_slot] =
			(usec_time+ourstats.unprocess_overhead[unprocess_slot])>> 1;
		unprocess_slot = (unprocess_slot+1)%BAND_ARRAY;
		break;
	}
}

/* What's the traffic like on this entry? */
void
packet_stats(struct sk_buff *skb, struct lin_hitbox *hitme)
{
	struct timeval our_time;
	long packettime;
	int last_seat, our_seat, seat;

	if (!skb->len)
		return;
	fixed_gettimeofday(&our_time);
	last_seat = band_seat(hitme->stats.last_packet);
	our_seat = band_seat(our_time);
	packettime = timeval_diff(&our_time, &hitme->stats.last_packet);
#ifdef TIME_DEBUG
	/* Check for bogus time values */
	if (packettime < 0) {
			printk("nistnet: packet arrived in %ld usecs!\n", packettime);
			goto after_bandwidth;
		}
#endif
	hitme->stats.last_packet = our_time;
	/* compute bandwidth */
	if (packettime > BAND_ARRAY*MILLION) {
		/* too long since last packet; zero out everything */
		memset(hitme->stats.bandwidth_array, 0,
			BAND_ARRAY*sizeof(unsigned long));
		hitme->stats.seats_used = 1;
	} else if (last_seat != our_seat) {	/* zero out skipped intervals */
		for (seat = (last_seat+1)%BAND_ARRAY; seat != our_seat;
			seat = (seat+1)%BAND_ARRAY) {
				hitme->stats.bandwidth_array[seat] = 0;
		}
		/* plus get this one! */
		hitme->stats.bandwidth_array[our_seat] = 0;
		if (hitme->stats.seats_used < BAND_ARRAY)
			++hitme->stats.seats_used;
	}
	hitme->stats.bandwidth_array[our_seat] += skb->len;

#ifdef TIME_DEBUG
after_bandwidth:
#endif
	hitme->stats.current_size = skb->len;
	hitme->stats.bytes_sent += skb->len;
	return;
}

/* Receive packet interception */
static struct packet_type *ippt;

static struct packet_type ourpt;

/* Resume processing of a delayed packet */
void
runpacket(struct fast_timer_list *info)
{
	struct hit_packetinfo *hpi = (struct hit_packetinfo *)info->data;
unsigned long pre_flags;

LinLock();
	packet_stats(hpi->skb, hpi->hitme);
LinUnlock();
	restore_flags(hpi->flags);
	(void) ippt->func(hpi->skb, hpi->dev, ippt);
LinLock();
if (!hpi->hitme->stats.qlen) {
	BREAKPOINT();
}
	--hpi->hitme->stats.qlen;
	fast_free(info);
LinUnlock();
	MOD_DEC_USE_COUNT;
}

/* This is the (slightly modified) DRD algorithm for dropping */

/* probability factors * PROBFACTOR */
static unsigned int drdtable[] = {	/* constant "ramp up" */
	6554,
	9830,
	13107,
	16384,
	19661,
	22938,
	26214,
	29491,
	32768,
	36045,
	39322,
	42598,
	45875,
	49152,
	52429,
	55706,
	58982,
	62259,
	65535
};
#define DRDLIMIT (sizeof(drdtable)/sizeof(int))

/* dropprob - compute probability of dropping packet.  If we are using
 * DRD, use its table (adjusted for the queue length parameters we're
 * using), otherwise use a constant drop probability (which may be 0).
 * If both DRD and constant drop are specified, we use the DRD probability
 * in preference to the constant drop if the former is non-zero.
 */
int
dropprob(struct lin_hitbox *hitme, int *use_drd, int *use_ecn)
{
	*use_drd=0;
	*use_ecn=0;
	if (hitme->hit_entry.drd_max) {	/* using DRD */
	    unsigned int stretch = DRDLIMIT/(hitme->hit_entry.drd_max-hitme->hit_entry.drd_min);

	    if (hitme->stats.qlen < hitme->hit_entry.drd_min) {
		return hitme->hit_entry.p_drop;
	    }
	    *use_drd=1;
	    if (hitme->stats.qlen >= hitme->hit_entry.drd_max)
		return drdtable[DRDLIMIT-1];
#ifdef CONFIG_ECN
	    /* If using DRD, check whether the queue length is still
	     * below the ECN limit.  If so, the packet can be marked
	     * with the ECN bit rather than dropped.
	     */
	    else if (hitme->stats.qlen <= hitme->hit_entry.drd_congestion)
	    	*use_ecn=1;
#endif /* CONFIG_ECH */
	    return drdtable[stretch *
			(hitme->stats.qlen-hitme->hit_entry.drd_min)];
	} else {
		return hitme->hit_entry.p_drop;
	}
}

#ifdef CONFIG_ECN
/* ecn_skb - mark a packet for explicit congestion notification, if supported.
 */
int
ecn_skb(struct sk_buff *skb)
{
	struct iphdr *iph;

	/* Get the ip header */
	iph = skb->h.iph;
	/* Check if ecn enabled */
	if (!(iph->tos & ECN_CAPABLE))
		return -1;
	/* Munge bit, if not already munged */
	if (!(iph->tos & ECN_NOTED)) {
		unsigned long checksum;

		iph->tos |= ECN_NOTED;
		/* Adjust checksum to account for munged bit */
		/* ip checksum is 1's complement in network byte order... */
		checksum = iph->check - htons(ECN_NOTED);
		checksum += checksum >>16;	/* catch carry */
		iph->check = checksum;
	}
	return 0;
}
#endif /* CONFIG_ECN */

/* Determine the amount of time to delay a packet.  This is the maximum
 * of two quantities:
 *	1. Probabilistic packet delay time
 *	2. Bandwidth-limitation delay time
 *
 * Question: Should we take probabilistic delay into account in determining
 * bandwidth consumption?  Answer: This complicates things a little too much.
 * Our model is that bandwidth throttling happens first at one virtual
 * choke point, then packets may get independently delayed at some later
 * point.  This can result in packets getting bunched up, so the stated
 * bandwidth limitation is actually exceeded at some point.
 *
 * The problem with taking the delay into account is that the simplest
 * way of doing so would remove any possibility of reordering packets --
 * each packet could not be sent any sooner than its predecessor.  It
 * thus seems more useful in terms of network effects to do things the
 * way they are here.
 */
int
packet_delay(struct sk_buff *skb, struct lin_hitbox *hitme)
{
	int probdelay=0, bandwidthdelay=0;

	/* Figure probabilistic delay */
	if (hitme->hit_entry.delsigma)
		probdelay = tabledist(hitme->hit_entry.delay,
			hitme->hit_entry.delsigma);
	else
		probdelay = hitme->hit_entry.delay;

	/* Figure bandwidth-limitation delay */
	if (hitme->hit_entry.bandwidth) {
		struct timeval our_time;
		long packettime;

		fixed_gettimeofday(&our_time);
		/* We can't send until queued packets have been taken care of */
		bandwidthdelay = timeval_minidiff(&hitme->stats.next_packet,
					&our_time);
		if (bandwidthdelay < 0) {
			bandwidthdelay = 0;
			hitme->stats.next_packet = our_time;
		}
		/* Now determine how much time this packet will take in
		 * usec, in order to schedule the following one.
		 */
		/* skb->len can sometimes be too big (with other junk)(?) */
		packettime = (long)skb->len*(MILLION/hitme->hit_entry.bandwidth) +
				((long)skb->len*(MILLION%hitme->hit_entry.bandwidth)
					+ hitme->hit_entry.bandwidth/2)/hitme->hit_entry.bandwidth;
		/* Quick defensive hack: even at 1 byte/second, a packet
		 * shouldn't take longer than MTU seconds!
		 */
		if (packettime < 0 || packettime > 1500*MILLION) {
#ifdef TIME_DEBUG
			printk("nistnet: wacky packettime of %ld, with length %ld and bandwidth %d\n",
				packettime, skb->len, hitme->hit_entry.bandwidth);
#endif
		} else {
			timeval_add(&hitme->stats.next_packet, packettime);
		}
#ifdef DEBUG
{
static int ticker;

++ticker;
if (!(ticker&0x3f)) {
	printk("nistnet: packet size %ld delay %ld usec\n",
		skb->len, packettime);
printk("nistnet: current time is %d.%06d, will send at %d.%06d\n",
		our_time.tv_sec, our_time.tv_usec,
		hitme->stats.next_packet.tv_sec, hitme->stats.next_packet.tv_usec);
}
}
#endif
	}


	return probdelay > bandwidthdelay ? probdelay : bandwidthdelay;
}

int
default_munger(struct sk_buff *skb, struct device *dev, struct packet_type *ptype, struct lin_hitbox *hitme)
{
	return 1;
}

packet_munger other_munge = default_munger;

void addmunge(packet_munger munger)
{
	other_munge = munger;
}

void rmmunge(packet_munger munger)
{
	if (munger == other_munge)
		other_munge = default_munger;
}

#define munge_finish()	{LinUnlock(); if (skb2) (void) rcv_packet_munge(skb2, dev, ptype);/* recursively process dup */ }

int
rcv_packet_munge(struct sk_buff *skb, struct device *dev, struct packet_type *ptype)
{
	unsigned long pre_flags;

LinLock();
	if (domunge) {
		int maybedrop, use_drd, use_ecn, delaytime;
		struct lin_hitbox *hitme;
		struct fast_timer_list *screamer;
		struct hit_packetinfo *hpi;
		struct sk_buff *skb2=NULL;
		int ret;

		global_stats(STATS_START);
		hitme = findhit(skb->h.iph->saddr, skb->h.iph->daddr);
		ret = (*other_munge)(skb, dev, ippt, hitme);
		if (ret <= 0) {
			global_stats(STATS_PROCESS); /* well, sort of */
LinUnlock();
			return ret;
		}
		if (!hitme) {	/* not intercepting */
			global_stats(STATS_UNPROCESS);
LinUnlock();
			return ippt->func(skb, dev, ippt);
		}

		/* Assume we will queue until we find otherwise */
		++hitme->stats.qlen;

		/* See if we're going to drop the packet */
		if ((maybedrop = dropprob(hitme, &use_drd, &use_ecn))) {
			int value = (random()&0xffff);

			if (value < maybedrop) {
#ifdef CONFIG_ECN
				/* ecn behavior: mark packet, don't drop */
				if (use_ecn && ecn_skb(skb) == 0) {
					++hitme->stats.drd_ecns;
				} else {
#endif /* CONFIG_ECN */
					kfree_skb(skb, FREE_WRITE);
					--hitme->stats.qlen;
					++hitme->stats.n_drops;
					if (use_drd)
						++hitme->stats.drd_drops;
					else
						++hitme->stats.rand_drops;
					global_stats(STATS_PROCESS);
LinUnlock();
					return 0;
#ifdef CONFIG_ECN
				}
#endif /* CONFIG_ECN */
			}
		}

		/* See if we're going to duplicate the packet.  Here,
		 * we just do fixed probability.
		 */
		if (hitme->hit_entry.p_dup) {
			int value = (random()&0xffff);
			
			if (value < hitme->hit_entry.p_dup) {	/* you get a new sister! */

				++hitme->stats.dups;
				skb2 = skb_copy(skb, GFP_ATOMIC);
			}
		}

		/* Now see if we're going to delay the packet */
		if (!(delaytime = packet_delay(skb, hitme))) { /* no delay */
			--hitme->stats.qlen;
			packet_stats(skb, hitme);

			global_stats(STATS_PROCESS);
munge_finish();
			return ippt->func(skb, dev, ippt);
		}

		screamer = fast_alloc(GFP_ATOMIC);
		/* If we can't allocate, punt! */
		if (!screamer) {
			--hitme->stats.qlen;
			++hitme->stats.mem_drops;
			packet_stats(skb, hitme);
			global_stats(STATS_PROCESS);
munge_finish();
			return ippt->func(skb, dev, ippt);
		}
		hpi = (struct hit_packetinfo *)screamer->data;

		init_fast_timer(screamer);
		hpi->skb = skb;
		hpi->dev = dev;
		hpi->hitme = hitme;
		hpi->flags = pre_flags;

		/* Schedule something to happen in a little while */
		screamer->expires = delaytime;
		MOD_INC_USE_COUNT;
		add_fast_timer(screamer);
		global_stats(STATS_PROCESS);
munge_finish();
		return 0;
	} else {
LinUnlock();
		return ippt->func(skb, dev, ippt);
	}
}

void
grab_ip_rcv(void)
{
	ippt = dev_find_pack(htons(ETH_P_IP));
	if (ippt) {
		ourpt = *ippt;
		ourpt.func = rcv_packet_munge;
		dev_remove_pack(ippt);
		dev_add_pack(&ourpt);
	}
}

void
release_ip_rcv(void)
{
	if (ippt) {
		dev_remove_pack(&ourpt);
		dev_add_pack(ippt);
		ippt = NULL;
	}
}

#ifdef MODULE
int
init_module( void)
#else
int
nistnet_init(void)
#endif
{
	void export_nistnet_symbols(void);

	printk("nistnet: init_module called\n");
	if (register_chrdev(major, "hw", &hw_fops)) {
		printk("nistnet: register_chrdev failed: goodbye world :-(\n");
		return -EIO;
	}
	else
		printk("nistnet: Hello, world\n");

	fast_fill();
	memset(&ourstats, 0, sizeof(ourstats));
	grab_ip_rcv();
	export_nistnet_symbols();
	return 0;
}

#ifdef MODULE
void
cleanup_module(void)
{
	printk("nistnet: cleanup_module called\n");

	if (unregister_chrdev(major, "hw") != 0)
		printk("nistnet: cleanup_module failed\n");
	else
		printk("nistnet: cleanup_module succeeded\n");
	domunge = 0;
	release_ip_rcv();
	freehittable();
	fast_empty();
}
#endif

/* We allocate 1024 slots at startup, then allow for extra bunches
 * of 64 at a time to be allocated if needed.  (We keep the extra
 * allocations small, since they are done at interrupt time, from
 * presumably precious locked-down kernel buffers.)
 *
 * Hence, the initial memory requirement is around 36K, while the
 * maximum allowed usage is on the order of 616K (17344 packets),
 * not counting the space used up by all those extra sk_buffs hanging
 * around.
 */
#define FAST_RESERVE 	1024
#define FAST_EMERGENCY	64
#define FAST_MAX	256

struct fast_timer_list *bigfastspace[FAST_MAX], *fast_stack;
struct hit_packetinfo *bighpispace[FAST_MAX];
int extra_count = 0;

void
fast_fill(void)
{
	struct fast_timer_list *newfast, *fastspace;
	struct hit_packetinfo *newhpi, *hpispace;
	int i, limit;

	if (!extra_count) {	/* First time, allocate a few pages */
		limit = FAST_RESERVE;
		fastspace = (struct fast_timer_list *)
			vmalloc(sizeof(struct fast_timer_list)*limit);
		if (!fastspace)
			return;
		hpispace = (struct hit_packetinfo *)
			vmalloc(sizeof(struct hit_packetinfo)*limit);
		if (!hpispace) {
			vfree(fastspace);
			return;
		}
	} else if (extra_count < FAST_MAX) {	/* subsequent times, go for fairly small chunks */
		limit = FAST_EMERGENCY;
		fastspace = (struct fast_timer_list *)
			kmalloc(sizeof(struct fast_timer_list)*limit, GFP_ATOMIC);
		if (!fastspace)
			return;
		hpispace = (struct hit_packetinfo *)
			kmalloc(sizeof(struct hit_packetinfo)*limit, GFP_ATOMIC);
		if (!hpispace) {
			kfree_s(fastspace, sizeof(struct fast_timer_list)*limit);
			return;
		}
	} else {			/* somebody got too greedy */
		return;
	}
	bigfastspace[extra_count] = fastspace;
	bighpispace[extra_count] = hpispace;
	++extra_count;

	for (i=0; i < limit; ++i) {
		newfast = fastspace+i;
		newhpi = hpispace+i;
		newfast->data = (unsigned long) newhpi;
		newfast->function = runpacket;
		newfast->next = fast_stack;
		fast_stack = newfast;
	}
}

void
fast_empty(void)
{
	int i;

	vfree((void *)bigfastspace[0]);
	vfree((void *)bighpispace[0]);
	for (i = 1; i < extra_count; ++i) {
		kfree_s(bigfastspace[i], sizeof(struct fast_timer_list)*limit);
		kfree_s(bighpispace[i], sizeof(struct hit_packetinfo)*limit);
	}
	extra_count = 0;
	fast_stack = NULL;
	memset((void *)bigfastspace, 0, sizeof(bigfastspace));
	memset((void *)bighpispace, 0, sizeof(bighpispace));
}

struct fast_timer_list *
fast_alloc(int how)
{
	struct fast_timer_list *answer = fast_stack;

	if (answer) {
		fast_stack = answer->next;
		return answer;
	}
	fast_fill();
	answer = fast_stack;
	if (answer)
		fast_stack = answer->next;
	return answer;
}

void
fast_free(struct fast_timer_list *done)
{
	done->next = fast_stack;
	fast_stack = done;
}

/* Export interfaces */
static struct symbol_table nistnet_syms = {
#include <linux/symtab_begin.h>

	X(addmunge),
	X(rmmunge),

#include <linux/symtab_end.h>
};

void export_nistnet_symbols(void)
{
	register_symtab(&nistnet_syms);
}
